/*
 * Universal Algebra Calculator
 * Copyright (C) 2001 Ralph Freese (ralph@math.hawaii.edu)
 *                    Emil W. Kiss (ewkiss@cs.elte.hu)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * The project's pages are at
 *     http://www.cs.elte.hu/~ewkiss/software/uaprog
 *     http://www.math.hawaii.edu/~ralph/software/uaprog
 */

package uacalc;

import java.awt.*;
import javax.swing.*;
import javax.swing.text.*;

/**
 * This class makes it possible to display progress information and
 * the results of calculations. It is the contents of the main
 * {@link uacalc.UACalc} window. It is a vertically split pane.
 * The top part (the results area) is {@link #ResultLog}, 
 * the bottom part (the progress area) is {@link #ProgressLog}.
 */
public class LoggingPane extends JSplitPane {

    public LoggingPane() {
	super(JSplitPane.VERTICAL_SPLIT);

	ResultLog = new LogArea(
	    defaultResultLogRows,
	    defaultNumColumns,
	    defaultResultLogMaxLines);

	ProgressLog = new LogArea(
	    defaultProgressLogRows,
	    defaultNumColumns,
	    defaultProgressLogMaxLines);

	ResultLog.setBorder(BorderFactory.createCompoundBorder(
	    BorderFactory.createTitledBorder("Results"),
	    BorderFactory.createEtchedBorder()));

	ProgressLog.setBorder(BorderFactory.createCompoundBorder(
	    BorderFactory.createTitledBorder("Progress"),
	    BorderFactory.createEtchedBorder()));

	setTopComponent(ResultLog);
	setBottomComponent(ProgressLog);
	setOneTouchExpandable(true);
	setContinuousLayout(true);
	setResizeWeight(2.0/3.0);
	updatePreferredSize();
    }

/**
 * Calculate and set the preferred size of this pane.
 * This is necessary after font changes, for example, to
 * make the horizontal scrollbars disappear.
 */
    public void updatePreferredSize () {

	Insets ins = ResultLog.getBorder().
	    getBorderInsets(ResultLog);
	Dimension d = new Dimension
	    ((int)ResultLog.getPreferredSize().width +
	     (int)ResultLog.getVerticalScrollBar().
	     getMinimumSize().getWidth() +
	     ins.left + ins.right,
	     (int)getPreferredSize().height);
	setPreferredSize(d);
    }

/**
 * Append text to the results area {@link #ResultLog}.
 *
 * @param text The text to be appended.
 * @see LogArea#limitedAppend
 */
    public void ResultLogAppend(String s) {
	ResultLog.limitedAppend(s);
    }

/**
 * Append text to the progress area {@link #ProgressLog}.
 *
 * @param text The text to be appended.
 * @see LogArea#limitedAppend
 */
    public void ProgressLogAppend(String text) {
	ProgressLog.limitedAppend(text);
    }

/**
 * The area to display results.
 */
    private LogArea ResultLog;
/**
 * The area to display progress information and errors.
 */
    private LogArea ProgressLog;
/**
 * Initial number of rows in the results area {@link #ResultLog}.
 */
    private int defaultResultLogRows = 20;
/**
 * Initial number of rows in the progress area {@link #ProgressLog}.
 */
    private int defaultProgressLogRows = 10;
/**
 * Initial number of columns in the progress and results areas
 * {@link #ProgressLog} and {@link #ResultLog}.
 */
    private int defaultNumColumns = 75;
/**
 * Initial number of rows to remember in the results area {@link #ResultLog}.
 */
    private int defaultResultLogMaxLines = 500;
/**
 * Initial number of rows to remember in the progress area {@link #ProgressLog}.
 */
    private int defaultProgressLogMaxLines = 500;

}

/**
 * Class to display textual information that is continously
 * produced during computations. It is a JTextArea enclosed in scroll bars.
 * Remembers only a given number of lines.
 */
class LogArea extends JScrollPane {

/**
 * Creates a LogArea with given dimensions.
 * The area is not editable and not wrapped.
 *
 * @param rows Number of rows to display initially.
 * @param columns Number of columns to display initially.
 * @param maxLines Number of lines to remember.
 * The user can scroll back only this many lines.
 */
    LogArea(int rows, int columns, int maxLines ) {

	m_maxLines = maxLines;
	area = new JTextArea(rows, columns);
	area.setEditable(false);
	area.setLineWrap(false);
	setViewportView(area);
    }

/**
 * Append a string to this log area, and scroll to the end
 * if necessary. Only a given number of lines are kept
 * (see {@link #setMaxLines }),
 * the rest is discarded from the beginning.
 *
 * @param text The text to be appended.
 *
 */
    public void limitedAppend(String text) {

	int lines = area.getLineCount();
	area.append(text);
	int linesafter = area.getLineCount();
	if( lines  > m_maxLines) {
	    int end = 0;
	    try {
		end = area.getLineEndOffset(
		    lines - m_maxLines);
	    } 
	    catch(BadLocationException e) {
		System.out.println(e.toString());
	    }
	    area.replaceRange(null, 0, end);
	}
	if(linesafter > lines) scrollToEnd();
    }

/**
 * Set the number of lines to remember.
 *
 * @param lines The number of lines to remember.
 *
 */
    public void setMaxLines(int lines) {
	m_maxLines = lines;
    }

/**
 * Scroll to the end of the text displayed.
 */
    private void scrollToEnd() {
	area.setCaretPosition(area.getText().length());
    }

/**
 * Number of lines to remember.
 */
    private int m_maxLines;
/**
 * The text area within the scroll pane.
 */
    private JTextArea area;
}
