/*
 * Universal Algebra Calculator
 * Copyright (C) 2001 Ralph Freese (ralph@math.hawaii.edu)
 *                    Emil W. Kiss (ewkiss@cs.elte.hu)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * The project's pages are at
 *     http://www.cs.elte.hu/~ewkiss/software/uaprog
 *     http://www.math.hawaii.edu/~ralph/software/uaprog
 */

package uacalc;

import uacalc.alg.Algebra;
import uacalc.alg.algedit.AlgebraEditor;
import uacalc.util.Project;
import uacalc.util.ewkFileChooser;
import uacalc.config.Configure;
import uacalc.config.MonospacedMetalTheme;
import uacalc.config.BigFontMetalTheme;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import java.io.*;
import javax.swing.plaf.metal.*;

/**
 * Universal Algebra Calculator, main module.
 * Contains the main window, and its menus:
 * <ul>
 * <li>File (allow quitting the program)
 * <li>Options (at the moment, allow to change font size)
 * <li>Algebra (allows editing, deleting, viewing computed results).
 * </ul>
 * The fourth menu, Compute, is in {@link uacalc.Compute}.
 */
public class UACalc extends JFrame {

    public UACalc() {

        JMenuBar menuBar = new JMenuBar();
        setJMenuBar(menuBar);

//File Menu
        JMenu FileMenu = new JMenu("File");
	FileMenu.setMnemonic(KeyEvent.VK_F);
        menuBar.add(FileMenu);

        JMenuItem QuitMenuItem = new JMenuItem("Quit", KeyEvent.VK_Q);
	FileMenu.add(QuitMenuItem);
	QuitMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_Q, ActionEvent.CTRL_MASK));

	QuitMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnClose();
		}
	    });

//Options Menu
	JMenu OptionsMenu = new JMenu("Options");
	OptionsMenu.setMnemonic(KeyEvent.VK_O);
	menuBar.add(OptionsMenu);

	BigFontMenuItem = 
	    new JCheckBoxMenuItem("Big font", false);//unchecked
	BigFontMenuItem.setMnemonic(KeyEvent.VK_B);
	OptionsMenu.add(BigFontMenuItem);
	BigFontMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_B, ActionEvent.CTRL_MASK));

	BigFontMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnBigFont();
		}
	    });


//Algebra Menu
	JMenu AlgebraMenu = new JMenu("Algebras");
	AlgebraMenu.setMnemonic(KeyEvent.VK_A);
	menuBar.add(AlgebraMenu);

//Edit
        JMenuItem EditMenuItem = 
	    new JMenuItem("Edit/Create algebra", KeyEvent.VK_E);
	AlgebraMenu.add(EditMenuItem);
	EditMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_E, ActionEvent.CTRL_MASK));

	EditMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnEdit();
		}
	    });

//Delete
        JMenuItem DeleteMenuItem = new JMenuItem("Delete algebra", KeyEvent.VK_D);
	AlgebraMenu.add(DeleteMenuItem);
	DeleteMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_D, ActionEvent.CTRL_MASK));

	DeleteMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnDelete();
		}
	    });

//Show
        JMenuItem ShowMenuItem = new JMenuItem("Show results", KeyEvent.VK_H);
	AlgebraMenu.add(ShowMenuItem);
	ShowMenuItem.setAccelerator(
	    KeyStroke.getKeyStroke(KeyEvent.VK_H, ActionEvent.CTRL_MASK));

	ShowMenuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    OnShow();
		}
	    });

//Application closes when window is closed
	addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    OnClose();
		}
	    });

	new Compute(this, menuBar);
    }

/**
 * Callback for the Close menu, exits program.
 */
    private void OnClose() {
	System.exit(0);
    }

/**
 * Callback for the Algebras/Edit menu,
 * starts the {@link uacalc.alg.algedit.AlgebraEditor}.
 * 
 */
    private void OnEdit() {
	new AlgebraEditor(this);
    }

/**
 * Callback for the Algebras/Show menu,
 * displays attributes computed so far.
 * Currently, these are read from the .dat file.
 */
    private void OnShow() {
	String fn = ewkFileChooser.getFile(
	    MainFrame, 
	    Project.getAlgebraFilter(),
	    "Select algebra to show its attributes",
	    ewkFileChooser.OPEN);
	if(fn == null) return;

	Project fn_Project = new Project(fn);
	Algebra A = new Algebra(fn);
	if( !A.ReadAlgebra() ) return;

	Progress.ResultLogAppend(
	    "\n [ Computed attributes of algebra " + 
	    "\n" + fn + "\n"
	    + A.toString());
        try {
	    BufferedReader in = new BufferedReader(
		new FileReader(fn_Project.getName(Project.InfoType)));
	    String line;
	    while(true) {
		line = in.readLine();
		if( line == null ) break;
		Progress.ResultLogAppend(line + "\n");
	    }
            in.close();
        }
        catch(Exception e) {
            System.out.println(e.toString());
        }
    }

/**
 * Callback for the Algebras/delete menu,
 * deletes an algebra and its associated files
 * (which is called a {@link uacalc.util.Project}).
 */
    private void OnDelete() {

	//Get the NAME of the project to delete.
	String fn = 
	    ewkFileChooser.getFile(this, 
				   Project.getAlgebraFilter(),
				   "Select algebra to delete",
				   ewkFileChooser.DELETE);
	if(fn == null) return;
	int ans = JOptionPane.showConfirmDialog(
	    this, 
	    "Are you sure you want to delete algebra\n"
	    + fn + 
	    "\nand all associated files?",
	    "Confirm remove",
	    JOptionPane.YES_NO_OPTION,
	    JOptionPane.WARNING_MESSAGE);
	if(ans != JOptionPane.YES_OPTION) return;

	Project m_Project = new Project(fn);
	m_Project.DeleteAndBackup();
	//Deletes locked projects, too!
    }

/**
 * Callback for the Options/BigFont checkbox menu,
 * sets font size.
 */
    private void OnBigFont() {

	if( BigFontMenuItem.getState() ) {
	    Configure.setBigBorder();
	    MetalLookAndFeel.setCurrentTheme( bigTheme );
	} else {
	    Configure.setDefaultBorder();
	    MetalLookAndFeel.setCurrentTheme( defaultTheme );
	}
	SetLaF();
        SwingUtilities.updateComponentTreeUI(MainFrame);
	log.updatePreferredSize();
        MainFrame.pack();
    }

/**
 * Sets Metal Look and Feel.
 */
    private static void SetLaF() {

	try {
	    UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");
	} catch (Exception ex) {
	    System.out.println("Failed loading Metal");
	    System.out.println(ex);
	}
    }

/**
 * The starting point.
 * Sets look and feel, fontsize, prepares the main frame
 * with menus, and the {@link uacalc.LoggingPane}.
 */
    public static void main(String[] args) {

	Configure.setDefaultBorder();
	defaultTheme = new MonospacedMetalTheme();
	bigTheme = new BigFontMetalTheme();
	MetalLookAndFeel.setCurrentTheme( defaultTheme );
	SetLaF();
	new ewkFileChooser();
        MainFrame = new UACalc();
	log = new LoggingPane();
	MainFrame.getContentPane().add(log);
        MainFrame.setTitle("Universal Algebra Calculator");
	MainFrame.pack();
        MainFrame.setVisible(true);
    }

/**
 * Toggles font size.
 */
    private JCheckBoxMenuItem BigFontMenuItem;
/**
 * Main Frame, containing menus and the {@link uacalc.LoggingPane}.
 */
    private static UACalc MainFrame;
/**
 * Theme with big fonts.
 * @see uacalc.config.BigFontMetalTheme
 */
    private static MetalTheme bigTheme;
/**
 * Theme with normal (default) fonts. 
 * @see uacalc.config.MonospacedMetalTheme
 */
    private static MetalTheme defaultTheme;
/**
 * A split pane containing progress and result output.
 * @see Progress
 */
    public static LoggingPane log;

}

