package uacalc.alg;

import uacalc.alg.con.Partition;
import uacalc.util.ewkInputStream;
import uacalc.vectors.VectorList;

import uacalc.ex.BadAlgebraException;
import uacalc.ex.AlgebraWriteException;
import uacalc.ex.BadOperationException;
import uacalc.ex.OperationWriteException;
import uacalc.ex.BadVectorListException;

import java.math.*;
import java.util.*;
import java.io.*;

public class Algebra {
    
    public Algebra(String fn) {
	m_size = 1;
	m_name = new String(fn); //copied!
	Operations = new ArrayList();
    }

    public int getSize() {
	return m_size;
    }

    public void setSize(int size) {
	m_size = size;
    }

    public String getName() {
	return m_name;
    }

    public void setName(String fn) {
	m_name = new String(fn); //copied!
    }

    public void deleteOp(int n) {
	Operations.remove(n);
	int i;
	for(i=n; i< Operations.size(); i++) {
	    ((Operation)Operations.get(i)).setName(i);
	}
    }

    public void switchOp(int n, int m) {

	Operation f = (Operation)Operations.get(n);
	Operation g = (Operation)Operations.get(m);
	f.setName(m);
	g.setName(n);
	Operations.set(n, g);
	Operations.set(m, f);
    }

    public void addOp(int n, Operation f) {
	//We do NOT check if f.getSize() == m_size

	Operations.add(n, f);
	int i;
	for(i=n+1; i< Operations.size(); i++) {
	    ((Operation)Operations.get(i)).setName(i);
	}
    }

    public int getNumOp() {
	//This is not speed critical.
	return Operations.size();
    }

    public Iterator getOperationsIterator() {
	return(Operations.iterator());
    }

    public Operation getOp(int i) throws ArrayIndexOutOfBoundsException {
	//Should we handle this exception here??
	return (Operation)Operations.get(i);
    }


    public void Read(ewkInputStream fp) 
	throws BadAlgebraException, BadOperationException {
	//Reads algebra from an already open file.

	int s = fp.get_integer("\n\r");
	if( s < 0) {
	    throw new BadAlgebraException(
		" ! Error reading size of algebra (bad delimiter?).");
	}
	if( s <= 0 || s >= ALG_SIZE_LIMIT) {
	    throw new BadAlgebraException("Illegal algebra size " + s + 
		" (should be in [" + 1 + "," + (ALG_SIZE_LIMIT-1) + "]).");
	}
	m_size = s;
	int opno=0;
	Operation f;
	while(true) {
	    f = new Operation(m_size, opno);
	    if(! f.Read(fp)) break;
	    if(opno >= OP_NUM_LIMIT) {
		throw new BadAlgebraException(
		    "Too many operations, should be at most"
		    + OP_NUM_LIMIT + ".");

	    }
	    Operations.add(f);
	    opno++;
	}
    }

    public void Write(PrintStream fp) 
	throws AlgebraWriteException, OperationWriteException {
	//Writes algebra to an already open file.

	fp.println(m_size);
	if( fp.checkError() ) {
	    throw new AlgebraWriteException(
		" ! Error writing the size of the algebra.");
	}
	Iterator it = Operations.iterator();
	while( it.hasNext() ) {
	    ((Operation)it.next()).Write(fp);
	}
    }


    public String toString() {

	String out = "";

	out += "Algebra has " + m_size + " elements.\n";
	Iterator it = Operations.iterator();
	while( it.hasNext() ) {
	    out += ((Operation)it.next()).toString();
	}
	return(out);
    }

    public boolean ReadAlgebra() {
 
        try {
            ewkInputStream fp = new ewkInputStream(
                new FileInputStream(m_name));
            try {
                Read(fp);
            }
            catch(BadAlgebraException e) {
                System.out.println(e.toString());
                fp.close();
                return(false);
            }
            catch(BadOperationException e) {
                System.out.println(e.toString());
                fp.close();
                return(false);
            }
            fp.close();
        }
        catch(Exception e) {
            System.out.println(e.toString());
            return(false);
        }
        return(true);
    } 

    public Algebra factor(Partition part, String name) {
	Algebra F = new Algebra(name);
	F.setSize(part.getNumberOfBlocks());
	int i;
	Iterator it = Operations.iterator();
	while( it.hasNext() ) {
	    Operation f = (Operation)it.next();
	    F.Operations.add( f.factor(part, f.getName()) );
	}
	return F;
    }


    public String isSimilar(Algebra B) {
	int numOp = getNumOp();
	if( numOp != B.getNumOp() ) {
	    return "The number of operations should be " + 
		numOp + ".";
	}
	int i;
	for(i=0; i< numOp; i++) {
	    if( getOp(i).getArity() != B.getOp(i).getArity() ) {
		return "The arity of operation " +
		    B.getOp(i).getNameString() + " should be "
		    + getOp(i).getArity() + ".";
	    }
	}
	return null;
    }

    //It is assumed that the algebras are similar
    public static Algebra directProduct(
	ArrayList components, VectorList universe, String name) 
	throws BadAlgebraException, BadOperationException,
	BadVectorListException {

	if( components == null || components.size() == 0 ) {
	    throw new BadAlgebraException(
		" ! Error: the number of components is zero!");
	}
	Algebra firstA = (Algebra)components.get(0);
	int numOp = firstA.getNumOp();

	int algSize = universe.size();
	if( algSize >= ALG_SIZE_LIMIT )  {
	    throw new BadAlgebraException(
		" ! Error: " +
		"size should be at most" + (ALG_SIZE_LIMIT-1) + ".");
	}

	Algebra P = new Algebra(name);
	P.setSize(algSize);

	ArrayList opComponents;
	Operation f,g;
	int i;
	Iterator it;
	for(i=0; i<numOp; i++) {//prepare the i-th operation of the product
	    opComponents = new ArrayList(components.size());
	    it = components.iterator();
	    while( it.hasNext() ) {
		f = ((Algebra)it.next()).getOp(i);
		opComponents.add(f);
	    }
	    g = Operation.directProduct(opComponents, universe,
					firstA.getOp(i).getName());
	    P.Operations.add(g);
	}
	return P;
    }
  
//Max number of operations:
    public static final int OP_NUM_LIMIT = 10000;
    public static final int ALG_SIZE_LIMIT = 10000;
    public static final int MAX_ALG_SIZE_DIGITS = 4;
    public static final String MAX_ALG_SIZE_STRING = "00000000";

    private int m_size;
    private String m_name;
    private ArrayList Operations;
}
