package uacalc.alg.con;

import uacalc.alg.Algebra;
import uacalc.alg.Operation;
import uacalc.vectors.IntArray;
import uacalc.vectors.SimpleList;
import uacalc.Progress;

import uacalc.ex.ComputationStoppedException;

/* CongruenceLattice.java 2001/06/04 Ralph Freese */
/* 
 * Changes by RSF: 2001/06/18
 * I changed makePrincipals so that it so that equal principal congruences
 * are indentical. This will result in a big space saving im many cases.
 *
 * Changes by EWK: 2001/06/14
 * This class contains only the core functionality
 * of the calculation itself. No I/O, and no
 * "helper" data holders for I/O.
 * This class is meant to be subclassed.
 * ?? Isn't it faster to use the JI congruences instead of
 * the principals when generating all congruences?
 * I changed it so.
 */

import java.util.*;
import java.io.*;

/**
 * A class to represent the congruence lattice of an algebra with universe
 * the integers from 0 to n-1. This uses the very fast algorithms from
 * my unpublished paper <i>Computing Congruences Efficiently</i>
 * which is available at 
 * {@link <a http://www.math.hawaii.edu/~ralph/papers.html}.
 *
 * @author Ralph Freese
 * @version $Id:$
 */ 
public class ComputeCongruenceLattice {

  protected Algebra alg;
  protected int algSize;
  protected int numOps;

  protected BasicPartition zeroCong;
  protected BasicPartition oneCong;


//Data holders:
  /** 
   * A map from pairs [i,j] to the array representing Cg(i, j).
   */
  protected HashMap principalCongruencesLookup = null;
  protected ArrayList congruences = null;
  protected HashMap upperCoversMap = null;
//Helpers:
///////////////////////////
// rsf: adding lowerCoverOfJIs back and I'm changing everything 
//      from private to protect so it can be accessed from 
//      CongruenceLattice.
///////////////////////////
  protected ArrayList principalCongruences = null;
  protected ArrayList joinIrredCongruences = null;
  protected HashMap lowerCoverOfJIs = null;
  protected HashSet congruencesHash = null;
  protected ArrayList meetIrredCongruences = null;

/////////////
//CONSTRUCTOR

    public ComputeCongruenceLattice(Algebra alg) {
	this.alg = alg;
	algSize = alg.getSize();
	numOps = alg.getNumOp();
	zeroCong = BasicPartition.zero(algSize);
	oneCong = BasicPartition.one(algSize);
  }


//    public void makeCon() {
//	makePrincipals();
//	makeJICongruences();
//	makeCongruences();
//	makeUpperCovers();
//    }


  public void makePrincipals() {
    HashMap pcIdMap = new HashMap();  // to keep equal congruences identical
    principalCongruences = new ArrayList();
    congruences = new ArrayList();
    congruencesHash = new HashSet();
    principalCongruencesLookup = new HashMap();
    for (int i = 0; i < algSize - 1; i++) {
      for (int j = i + 1; j < algSize; j++) {
        BasicPartition partCong = makeCg(i, j);
        if (pcIdMap.get(partCong) == null) {
          pcIdMap.put(partCong, partCong);
        }
        else {
          partCong = (BasicPartition)pcIdMap.get(partCong);
        }
	principalCongruencesLookup.put(new IntArray(new int[] {i, j}),
				       partCong);
	if( !congruencesHash.contains(partCong)) {
          principalCongruences.add(partCong);
          congruences.add(partCong);
          congruencesHash.add(partCong);
        }
      }
    }
  }

  public void makeCongruences() 
      throws ComputationStoppedException  {
    //Assumes that makePrincipals() is already called.
    //EWK: I changed principalCongruences to joinIrredCongruences.

//EWK
    Progress.StopProgram = false;

    int k = 0;
//    Iterator it = principalCongruences.iterator();
    Iterator it = joinIrredCongruences.iterator();
    while (it.hasNext()) {
      k++;
      BasicPartition elem = (BasicPartition)it.next();
      int n = congruences.size();
      for (int i = k; i < n; i++) {
        Partition join = elem.join((Partition)congruences.get(i));
//EWK
	if (Progress.StopProgram ) {
	    throw new ComputationStoppedException();
	}
        if (!congruencesHash.contains(join)) {

//TEMPORARY!! Make normal progress report later.
	    int s = congruences.size();
	    if( s % 1000 == 0) {
		Progress.ProgressLogAppend(
		    "\n [ Number of congruences is already " + s);
	    }

          congruencesHash.add(join);
          congruences.add(join);
        }
      }
    }
    congruencesHash.add(zeroCong);
    congruences.add(0, zeroCong);
  }

// rsf: I'm putting back the stuff to calculate the lower covers of ji's
//      It's almost free. And I need to have them without calculating
//      the conlat.
  public void makeJICongruences() {
    joinIrredCongruences = new ArrayList();
    lowerCoverOfJIs = new HashMap();
    Iterator it = principalCongruences.iterator();
    while (it.hasNext()) {
      Iterator it2 = principalCongruences.iterator();
      Partition part = (BasicPartition)it.next();
      Partition join = null;
      while (it2.hasNext()) {
        BasicPartition part2 = (BasicPartition)it2.next();
        if (part2.leq(part) && (!part.equals(part2))) {
          if (join == null) {
            join = part2;
          } 
          else {
            join = join.join(part2);
          }
        }
      }
      if (join == null || (!part.equals(join))) {
        joinIrredCongruences.add(part);
        if (join == null) {
          lowerCoverOfJIs.put(part, zeroCong);
        }
        else {
          lowerCoverOfJIs.put(part, join);
        }
      }
    }
  }

  /**
   * Makes the upperCoversMap. Assumes joinIrredCongruences have been
   * made.
   */
  public void makeUpperCovers() {
    upperCoversMap = new HashMap();
    Iterator it = congruences.iterator();
    // at first we put all elements of the form a \join j where j is a
    // join irred not below a onto the upper covers of a. This includes
    // all upper covers and possibly more. Then delete those that aren't 
    // covers.
    while (it.hasNext()) {
      BasicPartition elem = (BasicPartition)it.next();
      HashSet hs = new HashSet();
      ArrayList covs = new ArrayList();
      Iterator it2 = joinIrredCongruences.iterator();
      while (it2.hasNext()) {
        BasicPartition ji = (BasicPartition)it2.next();
        if (!ji.leq(elem)) {
          Partition join = ji.join(elem);
          if (!hs.contains(join)) {
            hs.add(join);
            covs.add(join);
          }
        }
      }
      hs = null;
      // note we need to use a listIterator here so we can remove elements
      // as we go.
      it2 = covs.listIterator();
      while (it2.hasNext()) {
        BasicPartition e0 = (BasicPartition)it2.next();
        boolean notCover = false;
        Iterator it3 = covs.iterator();
        while (it3.hasNext()) {
          BasicPartition e1 = (BasicPartition)it3.next();
          if ((!e0.equals(e1)) && e1.leq(e0)) {
            notCover = true;
            break;
          }
        }
        if (notCover) it2.remove();
      }
      upperCoversMap.put(elem, covs);
    }
  }

  /**
   * This assumes a < b.
   */
  public BasicPartition makeCg(int a, int b) {
    int[] part = new int[algSize];
    for (int i = 0; i < algSize; i++ ) {
      part[i] = -1;
    }
    part[a] = -2;
    part[b] = a;
    SimpleList pairs = new SimpleList(new int[] {a, b});
    while (!pairs.isEmpty()) {
      //int[] pair = (int[])pairs.first();
      int x = ((int[])pairs.first())[0];
      int y = ((int[])pairs.first())[1];
      pairs = pairs.rest();
      for (int i = 0; i < numOps; i++) {
        Operation f = alg.getOp(i);
        int arity = f.getArity();
	int[] arg = new int[arity];
	int[] arg2 = arg;
        for (int index = 0; index < arity; index++) {
	    arg = arg2;
	    for (int k = 0; k < arity; k++ ) {
		arg[k] = 0;
	    }
          while (arg != null) {
            arg[index] = x;
            int r = BasicPartition.root(f.getValue(arg), part);
            arg[index] = y;
            int s = BasicPartition.root(f.getValue(arg), part);
            if (r != s) {
              BasicPartition.joinBlocks(r, s, part);
              pairs = pairs.cons(new int[] {r, s});
            }
            arg = incrementArg(arg, index);
          }
        }
      }
    }
    return new BasicPartition(part);
  }

  /**
   * Increment the constants in arg which are in all positions
   * except <code>index</code>.
   */
  private int[] incrementArg(int[] arg, int index) {
    int length = arg.length;
    int max = algSize - 1;
    if (length < 2) return null;
    for (int i = 0; i < length; i++) {
      if (i == index) continue;
      if (arg[i] < max) {
        arg[i]++;
        return arg;
      }
      arg[i] = 0;
    }
    return null;
  }


  /**
   * This finds a meet irreducible congruence which is maximal with
   * respect to being above <code>a</code> and not above <code>b</code>.
   * Note if <code>a</code> is covered by <code>b</code> then
   * <code>[a,b]</code> transposes up to <code>[m,m*]</code>.
   */
  public Partition findMeetIrred (Partition a, Partition b) {
    if (b.leq(a)) return null;
    Iterator it = meetIrredCongruences.iterator();
    while (it.hasNext()) {
      Partition m = (Partition)it.next();
      if (a.leq(m) && !b.leq(m)) a = m;
    }
    return a;
  }

  /**
   * This finds a join irreducible congruence which is minimal with
   * respect to being below <code>b</code> and not below <code>a</code>.
   * Note if <code>a</code> is covered by <code>b</code> then
   * <code>[a,b]</code> transposes down to <code>[j,j_*]</code>.
   * Note joinIrredCongruences must have been constructed.
   */
  public Partition findJoinIrred (Partition a, Partition b) {
    if (b.leq(a)) return null;
    Iterator it = joinIrredCongruences.iterator();
    while (it.hasNext()) {
      Partition ji = (Partition)it.next();
      if (ji.leq(b) && !ji.leq(a)) b = ji;
    }
    return b;
  }

  private void makeMICongruences() {
    meetIrredCongruences = new ArrayList();
    Iterator it = congruences.iterator();
    while (it.hasNext()) {
      BasicPartition elem = (BasicPartition)it.next();
      if (((Collection)upperCoversMap.get(elem)).size() == 1) {
	meetIrredCongruences.add(elem);
      }
    }
  }

  /**
   * A the list of join irreducible congruences. This may return 
   * <code>null</code>, for example if the congruence lattice was 
   * imported from a file.
   */
  public List getJoinIrreducibles() {
    return joinIrredCongruences;
  }

  public final Partition getZeroCongruencePartition() {
    return zeroCong;
  }

  public final Partition getOneCongruencePartition() {
    return oneCong;
  }

}
