package uacalc.util;

import uacalc.Progress;

import java.io.*;
import javax.swing.*;

public class Project {

    public Project(String name) {
	//Must not be called with name = null.
	saveName(name);
    } 

//Locking
    protected int getLockCount() {
	//Returns -1 on error.
	//If the lock file is missing, it is created
	//with lock count = 0, and a message is written.

	String lockName = getName(LockType);
	File f = new File(lockName);
	if( !f.exists() ) {
	    if( !setLockCount(0, "Cannot write to lock file:")) {
		return(-1);
	    }
	    System.out.println("Created missing lock file " + lockName);
	    return(0);
	}
	int count;
        try {
            ewkInputStream fp = new ewkInputStream(
                new FileInputStream(lockName));
	    count = fp.get_integer("\n\r");
	    if( count < 0) {
		System.out.println("Lock file format error in "
				   + lockName + ".");
		return(-1);
	    }
	    fp.close();
        }
        catch(Exception e) {
            System.out.println(
		"Reading lock count unsuccesful:\n" +
		e.toString());
            return(-1);
        }
        return(count);
    }

    private boolean setLockCount(int count, String message) {
	//Returns false on I/O error and prints "message".
	//Creates lock file if it does not exist.

	String lockName = getName(LockType);
	try {
	    FileOutputStream fout =
		new FileOutputStream(lockName);
	    PrintStream fp =
		new PrintStream(fout);
	    fp.println(count);
	    fp.close();
	    fout.close();
	}
	catch(Exception e) {
	    System.out.println(message + "\n" + e.toString());
	    return(false);
	}
	return(true);
    }

    public boolean lock(boolean flag) {

	String message = "Setting lock count unsuccesful:";
	int count = getLockCount();
	if (count < 0) return(false);
	if( flag ) { //Lock
	    count++;
	    return(setLockCount(count, message));
	}
	//Unlock
	if(count > 0) count--;
	else {
	    count = 0;
	    System.out.println(
		"Unlocking unlocked project" +
		getName(LockType) + ".");
	}
	return(setLockCount(count, message));
    }

//Create

    public boolean create(JFrame frame, String startText) {
	//Returns false on cancel and on error.

	if( !WantsOverwriteAlgebra(frame)) return(false);

	ewkGetTextDialog dialog = new ewkGetTextDialog(
	    frame,
	    "Here you can enter any text describing algebra " +
	    getName(AlgebraType),
	    "Algebra Information Editor", startText);

	String datName = getName(InfoType);
	try {
	    FileOutputStream fout =
		new FileOutputStream(datName);
	    PrintStream fp =
		new PrintStream(fout);
	    fp.println(dialog.getEnteredText());
	    fp.close();
	    fout.close();
	}
	catch(Exception e) {
	    System.out.println("Cannot write to Info file"+
		datName);
	    return(false);
	}
	//Create lock file:
	return(setLockCount(0, "Cannot write to lock file:"));
    }

    private boolean WantsOverwriteAlgebra(JFrame frame) {
	//This function is called, when the user is about to
	//create this Project. The user has the option
	//to confirm this, to change the name of the project,
	//or to cancel.
	//Returns false if user canceled, true otherwise.
	//Thus the return value `true' does not express whether
	//the name has changed, or whether a project was deleted.
	//However, on cancel the name has not changed.

	//First try the existing name
	boolean newName = false;
	String fn = getName(AlgebraType);
	while(true) {
	    File file = new File(fn);
	    if( !file.exists() ) {
		if( newName ) { //Change name
		    saveName(fn);
		}
		return(true);
	    }
	    if( getLockCount() != 0 ) {//Project is locked
		JOptionPane.showMessageDialog(
                    frame,
                    "Please, select another name.",
                    "File is locked",
                    JOptionPane.INFORMATION_MESSAGE);
	    } else {
		int ans = JOptionPane.showConfirmDialog(
		    frame,
		    "You are about to overwrite the existing algebra \n\n" +
		    fn +
		    ". \n\nIf you press \"YES\" now, then that algebra, \n" +
		    "and all associated calculations will be lost. \n\n" +
		    "If you press \"NO\", then you can select a new name.\n",
		    "Confirm delete",
		    JOptionPane.YES_NO_CANCEL_OPTION,
		    JOptionPane.WARNING_MESSAGE);
		if(ans == JOptionPane.CANCEL_OPTION) return(false);
		if(ans != JOptionPane.NO_OPTION) {
		    DeleteAndBackup();
		    if( newName ) { //Change name
			saveName(fn);
		    }
		    return(true);
		}
	    }
	    fn = ewkFileChooser.getFile(frame, 
					AlgebraFilter,
					"Name of the new algebra",
					ewkFileChooser.NEW);
	    if(fn == null) return(false);
	    if( !fn.equals(m_name) ) {
	    //The new name is not the same as the old one.
		newName = true;
	    }
	}
    }

//Delete

    public void DeleteAndBackup() {
	//Deletes locked projects, too!

	int i;
	for(i=0; i<NumBackupTypes; i++) {
	    String s = getName(BackupTypes[i]);
	    File f = new File(s);
	    File g = new File(appendExtension(s, BackupType));
	    g.delete();//FIRST delete ALL backups so that the
	               //set of backups is consistent.
	    if( !f.exists() ) continue;
	    if( !f.renameTo(g) ) f.delete();
	}

	for(i=0; i<NumNonBackupTypes; i++) {
	    String s = getName(NonBackupTypes[i]);
	    File f = new File(s);
	    f.delete();
	}
    }

//Append
    public boolean appendToLog(String text) {

	try {
            FileOutputStream fout =
                new FileOutputStream(getName(InfoType), true);
	    //Open for append
            PrintStream fp = new PrintStream(fout);
	    fp.print(text);
	    Progress.ResultLogAppend(text);
            fp.close();
            fout.close();
        }
        catch(Exception e) {
	    System.out.println(e.toString());
	    return(false);
        }
	return(true);
    }


//Extension stuff
    public String getName(int eType) {
	return(appendExtension(m_name, eType));
    }

    // rsf  the file name with ext and without path
    public String getShortName() {
      return (new File(m_name)).getName();
    }

    public int getType() {
	return(m_eType);
    }

    public boolean existsName(int eType) {
	File file = new File(getName(eType));
	return(file.exists());
    }

/////////////////////////////

    protected String getExtension(int type) {
	if(type >= NumExtensionTypes) return(null);
	return(Extensions[type]);
    }

    private int getExtensionType(String Extension) {
	if( Extension == null ) return(NoType);
	int i;
	for(i=0; i<NumExtensionTypes; i++) {
	    if( i == NoType ) continue;
	    if( Extension.compareTo(Extensions[i]) == 0 ) {
		return(i);
	    }
	}
	return(NoType);
    }

    private String appendExtension(String s, int eType) {
        if(eType == NoType) return(s);
        return FileNameExtension.appendExtension(s, getExtension(eType));
    }

    protected void saveName(String s) {
        //Sets m_name and eType
        //If s has a valid extension, it is stripped,
        //otherwise m_name = s.
 
	String extension = FileNameExtension.getExtension(s);
	m_name = FileNameExtension.stripExtension(s);
        if( extension == null ) {
            m_eType = NoType;
            return;
        }
        m_eType = getExtensionType(extension);
    }

    public static final SimpleFileFilter getAlgebraFilter() {
	return(AlgebraFilter);
    }


    protected String m_name;
    protected int m_eType;

    protected static final String[] Extensions =
    {null, "bak", "par", "log",
     "alg", "dat", "uni",
     "lat", "cg", "con", "pri", "prm",
     "mi", "cov", "typ", "min",
     "p1", "p1e", "rlr",
     "vlf", "br", "br1", "br2", "has", "lck" };
    private static final int NumExtensionTypes = Extensions.length;

    public static final int NoType                  =  0;
    public static final int BackupType              =  1;
    public static final int ParType                 =  2;
    public static final int LogType                 =  3;
    public static final int AlgebraType             =  4;
    public static final int InfoType                =  5;
    public static final int UniverseType            =  6;
    public static final int ConLatType              =  7;
    public static final int CongruenceType          =  8;
    public static final int ConListType             =  9;
    public static final int PrincipalConListType    = 10;
    public static final int PrincipalMapType        = 11;
    public static final int MeetIrredConListType    = 12;
    public static final int CoverListType           = 13;
    public static final int TypeListType            = 14;
    public static final int MinSetType              = 15;
    public static final int UnaryPolType            = 16;
    public static final int IdempotentUnaryPolType  = 17;
    public static final int RLRType                 = 18;
    public static final int VectorListType          = 19;
    public static final int BinRelType              = 20;
    public static final int BinRelType1             = 21;
    public static final int BinRelType2             = 22;
    public static final int ConHasseType            = 23;
    public static final int LockType                = 24;

    private static final int[] BackupTypes =
    {AlgebraType, InfoType, UniverseType,
     ConLatType, ConListType, PrincipalConListType, PrincipalMapType,
     MeetIrredConListType, CoverListType, TypeListType,
     UnaryPolType, IdempotentUnaryPolType, RLRType};
    private static final int NumBackupTypes = BackupTypes.length;
    private static final int[] NonBackupTypes =
        {ParType, LogType, CongruenceType, MinSetType, 
         VectorListType, BinRelType, BinRelType1, BinRelType2, ConHasseType,
         LockType};
    private static final int NumNonBackupTypes = NonBackupTypes.length;

    private static final SimpleFileFilter AlgebraFilter =
    new SimpleFileFilter(Extensions[AlgebraType], "Algebra Files");
}

