/* 
 * SimpleList.java	98/1/3 Ralph Freese
 */

package uacalc.vectors;


import java.util.*;
import java.lang.reflect.*;

/**
 * Simple Linked lists. Java's collection framework has LinkedList's but
 * I need to guarentee that there is a great deal of sharing to save
 * space (memory).
 * For example, the space required to hold all subsets of an 
 * <i>n</i> element set requires space proportional to 
 * (<i>n</i>/2) 2<i><sup>n</sup></i> using vectors, but only
 * 2<i><sup>n</sup></i> with linked lists.
 * <p>
 * This version just has an element and a pointer to the rest, which
 * is another LinkedList. This means push and pop not are supported
 * but rest does not have to make a new object.
 * <p>
 * The rest of the empty list is itself.
 * <p>
 * This class implements <code>java.util.List</code>.
 * <p>
 * <b>Efficiency.</b> size() takes time proportional to the size and
 * get(int i) takes time proportional to i. <b>Moral:</b> do not use 
 * for loops to iterator over the elements; use the iterator. The former
 * uses time proportional to <i>n</i><sup>2</sup> while the latter 
 * uses only <i>n</i>.
 *
 *
 */
public class SimpleList implements Cloneable, java.util.List, 
                                              java.io.Serializable {

  protected Object first;
  protected SimpleList rest;

  /**
   * The empty list is a class constant
   */
  public static EmptyList emptyList = new EmptyList();

  private SimpleList() {}

  /**
   * Constructs a list with only obj.
   * @param obj The Object
   */
  public SimpleList(Object obj) {
    first = obj;
    rest = emptyList;
  }

/**
 * Constructs a list with obj followed by list. The same as cons in Lisp.
 * @param obj The Object to be first.
 * @param list The List to be the rest.
 */
  public SimpleList(Object obj, SimpleList list) {
    first = obj;
    rest = list;
  }

  public static SimpleList makeList() {
    return emptyList;
  }

  public static SimpleList makeList(Object obj) {
    return new SimpleList(obj);
  }

  public boolean isEmpty() {
    return false;
  }

  /**
   * The size of the list. It is inefficient since it takes time 
   * proportional to the size. 
   */
  public int size() {
    if (isEmpty()) return 0;
    return 1 + rest.size();
  }

  public Object first() {
    return first;
  }

  public SimpleList rest() {
    return rest;
  }

  public SimpleList cons(Object obj) {
    return new SimpleList(obj, this);
  }

  public Enumeration elements() {
    return new EnumerationSimpleList();
  }

  public Iterator iterator() { 
    return new EnumerationSimpleList();
  }


/*
  public ReusableIterator getIterator() {
    return new EnumerationSimpleList();
  }
*/

  public Iterator getIterator() {
    return new EnumerationSimpleList();
  }

  public SimpleList copyList() {
    if (isEmpty()) { 
      return emptyList ; 
    } else {
      return new SimpleList (first, rest.copyList()) ;
    }
  }

  public Object clone() {
    return copyList() ;
  }

  /**
   * This corresponds to <code>(APPEND this lst)</code> in lisp.
   */ 
  public SimpleList append(SimpleList lst) {
    if (isEmpty()) return lst;
    return new SimpleList(first(),rest().append(lst));
  }
   
  public SimpleList reverse() {
    return reverse(emptyList);
  }

  /**
   * This is revappend in Common Lisp. It produces 
   * <code>
   * (APPEND (REVERSE this) lst) 
   * </code>
   */ 
  public SimpleList reverse(SimpleList lst) {
    if (isEmpty()) return lst;
    SimpleList ans = lst;
    Enumeration list = (EnumerationSimpleList)this.elements(); 
    while (list.hasMoreElements()) { 
      ans = new SimpleList(list.nextElement(), ans);
    }
    return ans;
  }

  /**
   * This should use StringBuffer to be more efficient. Change it if
   * we use it for anything other than debugging.
   */
  public String toString() { 
    String str = "("; 
    Enumeration list 
         = (EnumerationSimpleList)this.elements(); 
    while (list.hasMoreElements()) { 
      str += list.nextElement() ; 
      if (list.hasMoreElements()) { str += " "; }
    }
    str += ")"; 
    return str; 
  }

  // methods demanded by List interface, most not real
  // some of these should be implemented

  /**
   * This just throws an UnsupportedOperationException.
   */
  public void add(int index, Object elt) throws 
               UnsupportedOperationException, ClassCastException,
               IllegalArgumentException, IndexOutOfBoundsException {
    throw new UnsupportedOperationException();
  }

  /**
   * This just throws an UnsupportedOperationException.
   */
  public boolean add(Object elt) throws UnsupportedOperationException, 
               ClassCastException, IllegalArgumentException {
    throw new UnsupportedOperationException();
  }

  /**
   * This just throws an UnsupportedOperationException.
   */
  public boolean addAll(int index, Collection c) throws 
               UnsupportedOperationException, ClassCastException,
               IllegalArgumentException, IndexOutOfBoundsException {
System.out.println("addAll called with index = " + index);
    throw new 
        UnsupportedOperationException("SimpleList does not support addAll");
  }

  public boolean addAll(Collection c) throws UnsupportedOperationException, 
               ClassCastException, IllegalArgumentException {
System.out.println("addAll called");
    throw new 
        UnsupportedOperationException("SimpleList does not support addAll");
  }

  /**
   * This just throws an UnsupportedOperationException.
   */
  public void clear() throws UnsupportedOperationException {
    throw new UnsupportedOperationException();
  }

  public boolean contains(Object o) { 
    Iterator it = iterator();
    while(it.hasNext()) {
      if (o.equals(it.next())) return true;
    }
    return false;
  }

  public boolean containsAll(Collection c) {
    Iterator it = c.iterator();
    while(it.hasNext()) {
      if (!contains(it.next())) return false;
    }
    return true;
  }

/*
  public boolean equals(Object o) throws UnsupportedOperationException { 
    throw new UnsupportedOperationException();
  }
*/

/*
  public Object get(int i) throws UnsupportedOperationException { 
    throw new UnsupportedOperationException();
  }
*/

  public Object get(int i) throws IndexOutOfBoundsException {
    if (i < 0 || i >= size()) throw new IndexOutOfBoundsException();
    Iterator it = iterator();
    int j = 0;
    while(it.hasNext()) {
      if (j++ == i) return it.next();
      it.next();
    }
    return null;		// can't happen
  }

  public int indexOf(Object o) { 
    Iterator it = iterator();
    int j = 0;
    while(it.hasNext()) {
      if (o.equals(it.next())) return j;
      j++;
    }
    return -1;
  }

  public int lastIndexOf(Object o) { 
    Iterator it = iterator();
    int j = 0;
    int ans = -1;
    while(it.hasNext()) {
      if (o.equals(it.next())) ans = j;
      j++;
    }
    return ans;
  }

  public Object remove(int i) throws UnsupportedOperationException { 
    throw new UnsupportedOperationException();
  }

  public boolean remove(Object i) throws UnsupportedOperationException { 
    throw new UnsupportedOperationException();
  }

  public boolean removeAll(Collection i) throws UnsupportedOperationException { 
    throw new UnsupportedOperationException();
  }

  public boolean retainAll(Collection i) throws UnsupportedOperationException { 
    throw new UnsupportedOperationException();
  }

  public Object set(int i, Object o) throws IndexOutOfBoundsException { 
    if (i < 0 || i >= size()) throw new IndexOutOfBoundsException();
    SimpleList lst = this;
    for (int j = 0; j < i; j++) {
      lst = lst.rest();
    }
    Object ans = lst.first;
    lst.first = o;
    return ans;
  }

/*
  public Object[] toArray(Object[] a) throws UnsupportedOperationException { 
    throw new UnsupportedOperationException();
  }
*/
  
  public Object[] toArray(Object[] a) { 
    int size = size();
    int aSize = a.length;
    if (size > aSize) {
      Class c = a.getClass().getComponentType();
      a = (Object[])Array.newInstance(c, size);
    }
    int i = 0;
    Iterator it = iterator();
    while (it.hasNext()) {
      a[i] = it.next();
      i++;
    }
    for (int j = i; j < aSize; j++) {
      a[j] = null;
    }
    return a;
  }
  
  public Object[] toArray() { 
    //throw new UnsupportedOperationException();
    Iterator it = iterator();
    Object[] ans = new Object[size()];
    int i = 0;
    while (it.hasNext()) {
      ans[i] = it.next();
      i++;
    }
    return ans;
  }

  public java.util.List subList(int i, int j) {
    int k = 0;
    SimpleList ans = emptyList;
    Iterator it = iterator();
    while (it.hasNext()) {
      if (i <= k && k < j) ans = ans.cons(it.next());
    }
    return ans.reverse();
  }

  public ListIterator listIterator(int i) {
    return new ListIteratorSimpleList(i);
  }

  public ListIterator listIterator() {
    return new ListIteratorSimpleList();
  }

  
  private static class EmptyList extends SimpleList {

    public EmptyList() {
      first = null;
      rest = this;
    }

    //private static EmptyList emptyList = new SimpleList(null);

    //static {
      //emptyList.rest = emptyList;
    //}

    public boolean isEmpty() {
      return true;
    }

  }



  private class ListIteratorSimpleList implements ListIterator {
    private ArrayList alist;
    private ListIterator iter;

    ListIteratorSimpleList() {
      alist = new ArrayList(SimpleList.this);
      iter = alist.listIterator();
    }

    ListIteratorSimpleList(int i) {
      alist = new ArrayList(SimpleList.this);
      iter = alist.listIterator(i);
    }

    public boolean hasNext() {
      return iter.hasNext();
    }

    public Object next() {
      return iter.next();
    }

    public int nextIndex() {
      return iter.nextIndex();
    }

    public boolean hasPrevious() {
      return iter.hasPrevious();
    }

    public Object previous() {
      return iter.previous();
    }

    public int previousIndex() {
      return iter.previousIndex();
    }

    public void add(Object o) {
      throw new UnsupportedOperationException();
    }

    public void remove() {
      throw new UnsupportedOperationException();
    }

    public void set(Object o) {
      throw new UnsupportedOperationException();
    }

  }


  // was ReusableIterator
  private class EnumerationSimpleList implements Enumeration, Iterator {
    private SimpleList llist;

    EnumerationSimpleList() {
      llist = SimpleList.this;
    }

    public boolean hasMoreElements() {
      return (! (llist.isEmpty()));
    }

    public boolean hasNext() {
      return (! (llist.isEmpty()));
    }

    public Object nextElement() {
      Object obj = llist.first();
      llist = llist.rest();
      return obj;
    }

    public Object next() {
      Object obj = llist.first();
      llist = llist.rest();
      return obj;
    }

/*
    public void reinitialize() { llist = SimpleList.this; }
    public ReusableIterator regenerate() { 
      return SimpleList.this.getIterator(); 
    }
*/

    public void remove () {
      throw new UnsupportedOperationException();
    }

/*
    public ReusableIterator removeConflicts(Conflictable obj) { 
      //return copyList().getIterator(); 
      return removeConflictsAux(obj, SimpleList.this).getIterator(); 
    }
*/

  }

/*
 * Al's slow reverse for time testing.
 */

/*
  public SimpleList rev() {
    if (isEmpty()) {
      return this;
    } else {
      return this.rest.rev().app(new SimpleList(first));
    }
  }
      
  public SimpleList app(SimpleList List) {
    if (isEmpty()) {
      return List ;
    } else {
      return
        rev().rest.rev().app(new SimpleList(rev().first, List));
    }
  }

*/

  public static void main(String[] args) {
System.out.println("got here");
    int n;
    if (args.length != 0) {
      n = Integer.parseInt(args[0]);
    } else {
      n = 10;
    }
System.out.println("got here2");
    SimpleList foo = new SimpleList(new Integer(0)); 
    for(int i = 1; i < n; i++) {
      //foo = new SimpleList(new Integer(i), foo);
      foo = foo.cons(new Integer(i));
    }
System.out.println("got here");
SimpleList goo = emptyList.cons(new Integer(55));
System.out.println("goo is " + goo.toString());
    System.out.println("the list is " + foo.toString());
    //System.out.println("the size is " + foo.size());
    long time = System.currentTimeMillis();
    //System.out.println("the rev list is " + foo.rev().toString());
    System.out.println("the reverse list is " + foo.reverse().toString());
    System.out.println("list append list is " + foo.append(foo).toString());
    System.out.println("list revappend list is " + foo.reverse(foo).toString());
    time = System.currentTimeMillis() - time;
    System.out.println("Compute time is " + time);
System.out.println("first of emptyList is " + emptyList.first());
System.out.println("rest of emptyList is " + emptyList.rest());
System.out.println("emptyList is empty? " + emptyList.isEmpty());
System.out.println("foo is empty? " + foo.isEmpty());
System.out.println("the size of foo is " + foo.size());


System.out.println("\nfoo as an array is  " + foo.toArray().length);
ArrayList a = new ArrayList(foo);
//System.out.println("\nfoo as an array is  " + a);
//System.out.println("a[5] is  " + a[5]);
//Collections.shuffle(foo);
//System.out.println("\nfoo shuffled is " + foo);
//foo.set(5, new Integer(-1));
//System.out.println("\nfoo with new five is " + foo);
//System.out.println("\nfoo shuffled is " + foo);
System.out.println("\nfoo is " + foo);
// didn't work 
//Collections.sort(foo);
//System.out.println("\nfoo sorted is " + foo);
    
  }

}

  


